package de.uka.algo.GT3D;

import java.awt.AWTException;
import java.awt.BorderLayout;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.GraphicsConfiguration;
import java.awt.Point;
import java.awt.Robot;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.StringTokenizer;

import javax.media.j3d.BranchGroup;
import javax.media.j3d.Canvas3D;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.SwingConstants;
import javax.swing.Timer;
import javax.vecmath.Matrix3d;
import javax.vecmath.Vector2d;
import javax.vecmath.Vector3d;
import javax.vecmath.Vector3f;

import com.sun.j3d.utils.picking.PickCanvas;
import com.sun.j3d.utils.picking.PickResult;
import com.sun.j3d.utils.universe.SimpleUniverse;
import com.sun.j3d.utils.universe.ViewingPlatform;

public class VisualizationFramework extends JFrame implements ActionListener, KeyListener, MouseMotionListener, MouseListener, MouseWheelListener {
	private Canvas3D c3d;

//	private JButton go = new JButton("Start");

	private JButton cameraInfo = new JButton("Camera Info");

	private JButton invertMouseY = new JButton("Invert Mouse (y)");

	private JLabel messageLabel = new JLabel("");

	private JLabel coordLabel = new JLabel("");

	private int currentPressedKey;
	private Timer timer;

	private int timerMillis = 20;

	boolean recordOn = false;

	boolean playOn = false;

	int tourCounter = 0;

	String navFile = "autopilot.txt";

	BufferedWriter statsWrite;

	BufferedReader statsRead;

	private TransformGroup flyingCamera = new TransformGroup();

	private Transform3D moveCamera = new Transform3D();

	private ViewingPlatform vPlatform;

	private PickCanvas pickCanvas;

	private BranchGroup scene;

	private int aimTolerance = 22;

	private double lookSensitivity = 0.003;

	private double knueppelSensitivity = 0.02;

	private double moveSensitivity = 0.3;

	private double keyRollFactor = 50.0;

	private double mWheelRollFactor = 400.0;

	private boolean steuerKnueppelMode = true;

	private Matrix3d camHeading = new Matrix3d(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);
	private Vector3d camPosition = new Vector3d(0.0, 0.0, 20.0);

	private Matrix3d rotMatX = new Matrix3d(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0); 
	// these will be modified according to mouse movement,
	// then multiplied with camHeading

	private Matrix3d rotMatY = new Matrix3d(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);

	private Matrix3d rotMatZ = new Matrix3d(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);

	private Matrix3d unitMatrix3D = new Matrix3d(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);

	private Matrix3d rotMatYtoMinusZ = new Matrix3d(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);

	private Matrix3d startRotation = new Matrix3d(1,0,0, 0,1,0, 0,0,1);//(0.9986783280557308, 0.0035689514255241466, -0.05127240639502675, 0.001182416936916051, 0.9957267237888925, 0.09234117945324663, 0.051382866424537106, -0.09227976006878073, 0.9944064797252914);

	private Vector3d dazuPosition = new Vector3d(0.0, 0.0, 0.0);

	private Vector3d startPosition = new Vector3d(0,0,0);//(-7.59, 6.89, 86.30);

	private Vector3f tempRotAxis = new Vector3f(0.0f, 1.0f, 0.0f);

	double mX = 0.0; // for turning calculation
	double cosmX = 0.0;
	double sinmX = 0.0;

	double mY = 0.0;
	double cosmY = 0.0;
	double sinmY = 0.0;

	double mZ = 0.0;
	double cosmZ = 0.0;
	double sinmZ = 0.0;

	double refX = 0.0; // reference (last) coordinate for mouse movement
	// detection
	double refY = 0.0;

	Point canvasTopLeft;

	double Ysign = -1.0; // -1 means invertY, 1 means arcade

	static int resolutionX = 1000; // MainFrame Resolution
	static int resolutionY = 700;

//	boolean nodeTextureOn = true;

//  TODO: raus?
//	int xDivision = 30;
//	int yDivision = 1;

	/**
	 * Used for moving the cursor to the center when reset key is pressed
	 */
	Robot theRobot;



	/**
	 * 
	 * 
	 * @param kl A key listener that will get all key events for the window
	 * @throws AWTException 
	 */
	public VisualizationFramework(KeyListener kl) throws AWTException {
		setTitle("Navigable Universe");
		setSize(new Dimension(600, 400));
		setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
		// TODO: schließt nicht beim Schließen
		setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		setLayout(new BorderLayout());

		GraphicsConfiguration config = SimpleUniverse.getPreferredConfiguration();
		c3d = new Canvas3D(config);
		canvasTopLeft = c3d.getLocationOnScreen();
		c3d.setDoubleBufferEnable(true);
		c3d.addMouseMotionListener(this);
		c3d.addKeyListener(this);
		c3d.addKeyListener(kl);
		c3d.addMouseListener(this);
		c3d.addMouseWheelListener(this);
		c3d.setMinimumSize(new Dimension(600, 400));
		addComponentListener(new ComponentAdapter() {

			@Override
			public void componentResized(ComponentEvent e) {
				super.componentResized(e);
				// Component c3d = e.getComponent();
				if (c3d == null)
					return;
				resolutionY = c3d.getHeight();
				resolutionX = c3d.getWidth();
				refX = canvasTopLeft.getX() + resolutionX / 2.0;
				refY = canvasTopLeft.getY() + resolutionY / 2.0;
			}

			@Override
			public void componentMoved(ComponentEvent e) {
				super.componentMoved(e);
				// Component c3d = e.getComponent();
				canvasTopLeft = c3d.getLocationOnScreen();
				refX = canvasTopLeft.getX() + resolutionX / 2.0;
				refY = canvasTopLeft.getY() + resolutionY / 2.0;
			}
		});

		add(c3d, BorderLayout.CENTER);

		// responsible for refresh rate etc.
		timer = new Timer(timerMillis, this);
		currentPressedKey = KeyEvent.VK_UNDEFINED;        
		// timer.start();


		// ------------------------   USER INTERFACE ------------------------------------
		// TODO: gscheites User Interface bauen
		JPanel buttons = new JPanel();
//		go.setPreferredSize(invertMouseY.getPreferredSize());
//		buttons.add(go);
		buttons.add(invertMouseY);

		messageLabel.setHorizontalAlignment(SwingConstants.CENTER);

		JPanel statusbar = new JPanel(new BorderLayout());
		statusbar.add(buttons, BorderLayout.LINE_START);
		statusbar.add(messageLabel, BorderLayout.CENTER);
		statusbar.add(coordLabel, BorderLayout.LINE_END);
		add(statusbar, BorderLayout.PAGE_END);

//		go.addActionListener(this);
//		go.addKeyListener(this);
//		go.addKeyListener(kl);
		invertMouseY.addActionListener(this);
		invertMouseY.addKeyListener(this);
		invertMouseY.addKeyListener(kl);
		if (steuerKnueppelMode) {
			lookSensitivity *= knueppelSensitivity;
		}


		// picking
		pickCanvas = new PickCanvas(c3d, scene);
		pickCanvas.setMode(PickCanvas.BOUNDS);
		
		theRobot = new Robot();
	}

	/**
	 * inserts a root node of a 3d scene into a freshly created universe.  
	 * 
	 * @param rootNode The root of the scene to insert.
	 */
	public void setRoot(BranchGroup rootNode){
		scene = rootNode;

		// create the universe
		SimpleUniverse universe = new SimpleUniverse(c3d);
		universe.addBranchGraph(scene);

		// set up moving camera
		vPlatform = universe.getViewingPlatform();
		vPlatform.setNominalViewingTransform();
		flyingCamera = vPlatform.getViewPlatformTransform();
		moveCamera.setTranslation(startPosition);
		moveCamera.setRotation(startRotation);
		flyingCamera.setTransform(moveCamera);
		universe.getViewer().getView().setBackClipDistance(100.0);
	}
	
	
	/**
	 * resets the camera heading
	 */
	private void rotMatReset() {
		rotMatX.set(unitMatrix3D);
		rotMatY.set(unitMatrix3D);
		rotMatZ.set(unitMatrix3D);
	}

	public void keyTyped(KeyEvent e) {
		// Invoked when a key has been typed.
		if (e.getKeyChar() == 'r') { // reset everything (position, heading,
			// movement)!!
			resetPosAndHeading();
			currentPressedKey = KeyEvent.VK_UNDEFINED;
		}
		else if (e.getKeyChar() == 'q') { // invert up-down-look,
			// arcade<->simulator
			theRobot.mouseMove(((int) refX), ((int) refY) + 10);
			currentPressedKey = KeyEvent.VK_UNDEFINED;
		}
		else if (e.getKeyChar() == 'n') {
			printCameraStats();
			currentPressedKey = KeyEvent.VK_UNDEFINED;
		} 
		else if (e.getKeyChar() == 't') {
			tourPlay();
			tourCounter = 0;
			playOn = !playOn;
			try {
				if (playOn) {
					statsRead = new BufferedReader(new FileReader(navFile));
				} else
					statsRead.close();
			} catch (IOException ioe) {
			}
			currentPressedKey = KeyEvent.VK_UNDEFINED;
		}

		else if (e.getKeyChar() == 'z') {
			tourRecord();
			currentPressedKey = KeyEvent.VK_UNDEFINED;
		}

		else if (e.getKeyChar() == 'y') { // invert up-down-look,
			// arcade<->simulator
			Ysign *= -1.0;
			currentPressedKey = KeyEvent.VK_UNDEFINED;
		}

		else if ((e.getKeyChar() == java.awt.event.KeyEvent.VK_ENTER) && (!timer.isRunning())) {
			timer.start();
			resetPosAndHeading();
			currentPressedKey = KeyEvent.VK_UNDEFINED;
		} 
	}


	
	public void keyPressed(KeyEvent e) {
		// Invoked when a key has been pressed.
		if (currentPressedKey == KeyEvent.VK_UNDEFINED) {
			currentPressedKey = e.getKeyCode();
		}
	}

	public void keyReleased(KeyEvent e) {
		if (currentPressedKey == e.getKeyCode())
			currentPressedKey = KeyEvent.VK_UNDEFINED;
	}

	
	/**
	 * interprets user input and performs calculations
	 * 
	 * @param keyNumber The code of the currently pressed key.
	 */
	public void keyEvaluation(int keyNumber) {
		// X-Axis (left - right(pos))
		if (keyNumber == KeyEvent.VK_S) {
			camHeading.getColumn(0, dazuPosition);
			dazuPosition.scale(-1.0 * moveSensitivity);
			camPosition.add(dazuPosition);
		} else if (keyNumber == KeyEvent.VK_F) {
			camHeading.getColumn(0, dazuPosition);
			dazuPosition.scale(moveSensitivity);
			camPosition.add(dazuPosition);
		}

		// Y-Axis (up(pos) - down)
		else if (keyNumber == java.awt.event.KeyEvent.VK_SPACE) {
			camHeading.getColumn(1, dazuPosition);
			dazuPosition.scale(-1.0 * moveSensitivity);
			camPosition.add(dazuPosition);
		} else if (keyNumber == KeyEvent.VK_LESS) {
			camHeading.getColumn(1, dazuPosition);
			dazuPosition.scale(moveSensitivity);
			camPosition.add(dazuPosition);
		}

		// Z-Axis (outOfScreen(pos) - intoScreen)
		else if (keyNumber == KeyEvent.VK_E) {// forward
			camHeading.getColumn(2, dazuPosition);
			dazuPosition.scale(-1.0 * moveSensitivity);
			camPosition.add(dazuPosition);
		} else if (keyNumber == KeyEvent.VK_D) {// backward
			camHeading.getColumn(2, dazuPosition);
			dazuPosition.scale(moveSensitivity);
			camPosition.add(dazuPosition);
		}

		// Turn Left and Right around camera-Y-axis
		else if (keyNumber == KeyEvent.VK_J) {// left
			cosmY = Math.cos(lookSensitivity * -1.0 * keyRollFactor); // Eigentlich
			sinmY = Math.sin(lookSensitivity * -1.0 * keyRollFactor); // nur
			rotMatY.setColumn(0, cosmY, 0.0, sinmY);
			rotMatY.setColumn(2, -sinmY, 0.0, cosmY);
			camHeading.mul(rotMatY);
			rotMatReset();
		} else if (keyNumber == KeyEvent.VK_L) { // KeyEvent.VK_KP_RIGHT geht irgendwie
			// ned
			cosmY = Math.cos(lookSensitivity * 1.0 * keyRollFactor);
			sinmY = Math.sin(lookSensitivity * 1.0 * keyRollFactor);
			rotMatY.setColumn(0, cosmY, 0.0, sinmY);
			rotMatY.setColumn(2, -sinmY, 0.0, cosmY);
			camHeading.mul(rotMatY);
			rotMatReset();
		}

		// Turn Up and Down around camera-X-axis
		else if (keyNumber == KeyEvent.VK_I) { // up
			cosmX = Math.cos(lookSensitivity * -1.0 * keyRollFactor);
			sinmX = Math.sin(lookSensitivity * -1.0 * keyRollFactor);
			rotMatX.setColumn(1, 0.0, cosmX, -sinmX);
			rotMatX.setColumn(2, 0.0, sinmX, cosmX);
			camHeading.mul(rotMatX);
			rotMatReset();
		} else if (keyNumber == KeyEvent.VK_K) { // down
			cosmX = Math.cos(lookSensitivity * 1.0 * keyRollFactor);
			sinmX = Math.sin(lookSensitivity * 1.0 * keyRollFactor);
			rotMatX.setColumn(1, 0.0, cosmX, -sinmX);
			rotMatX.setColumn(2, 0.0, sinmX, cosmX);
			camHeading.mul(rotMatX);
			rotMatReset();
		}

		// Roll Left and Right around camera-Z-axis
		else if (keyNumber == KeyEvent.VK_U) { // roll left
			cosmZ = Math.cos(lookSensitivity * 1.0 * keyRollFactor);
			sinmZ = Math.sin(lookSensitivity * 1.0 * keyRollFactor);
			rotMatZ.setColumn(0, cosmZ, sinmZ, 0.0);
			rotMatZ.setColumn(1, -sinmZ, cosmZ, 0.0);
			camHeading.mul(rotMatZ);
			rotMatReset();
		} else if (keyNumber == KeyEvent.VK_O) { // roll right
			cosmZ = Math.cos(lookSensitivity * -1.0 * keyRollFactor);
			sinmZ = Math.sin(lookSensitivity * -1.0 * keyRollFactor);
			rotMatZ.setColumn(0, cosmZ, sinmZ, 0.0);
			rotMatZ.setColumn(1, -sinmZ, cosmZ, 0.0);
			camHeading.mul(rotMatZ);
			rotMatReset();
		}
	}

	
	private void addStatsToNavFile() {
		try {
			statsWrite.newLine();
			statsWrite.newLine();
			statsWrite.append(camHeading.toString());
			statsWrite.append(camPosition.toString());
		} catch (IOException e) {
			System.err.println("HILFEEEEEEEEEEE");
		}
		;
	}

	private void initializeNavFile() {
		try {
			statsWrite = new BufferedWriter(new FileWriter(navFile));
		}
		// statsWrite.write("# This is a navigation file for a tour");
		catch (IOException e) {
			System.err.println("HILFEEEEEEEEEEE");
		}
		;
	}

	private void tourRecord() {
		recordOn = !recordOn;
		if (recordOn) {
			initializeNavFile();
		} else {
			try {
				statsWrite.close();
			} catch (IOException e) {
				System.err.println("HILFEEEEEEEEEEE");
			}
			;
		}
	}

	private void tourPlay() {
	}

	public void resetPosAndHeading() {
		// int x = resolutionX / 2;
		// int y = resolutionY / 2;
		// camHeading.set(startRotation);
		camHeading.set(startRotation);
		camPosition.set(startPosition);
		theRobot.mouseMove(((int) refX), ((int) refY) + 10);
		rotMatReset();
		// System.out.println(refX + " " + refY + 10);
	}

	public void actionPerformed(ActionEvent e) {
//		// start timer when button is pressed
//		if ((e.getSource() == go) && (!timer.isRunning())) {
//			timer.start();
//			resetPosAndHeading();
//		}
		
		// update camera heading etc. when the timer has fired
		if (e.getSource() == timer){
			if (playOn){
				// read next movie frame
				try {
					if (!getNextStepData()) {
						playOn = false;
						statsRead.close();
					}
				} catch (IOException e1) {
					System.err.println("HILFEEEEEEE! " + e1.getStackTrace());
				}
			}else{
				// update camera stats according to pressed keys...
				keyEvaluation(currentPressedKey);
				
				// .. and mouse position (die rotMat wird unabhängig vom Timer gerefresht)
				camHeading.mul(rotMatX);
				camHeading.mul(rotMatY);
				
				if (recordOn)
					addStatsToNavFile();
			}
			// do the move
			moveCamera.setRotation(camHeading);
			moveCamera.setTranslation(camPosition);
			flyingCamera.setTransform(moveCamera);
			coordLabel.setText("Position: " + printVector(camPosition, 2, 1) + "   Heading: " + printVector((new Vector3d(camHeading.m02, camHeading.m12, camHeading.m22)), 1, 2));
		}		
		if (e.getSource() == invertMouseY) {
			Ysign *= -1.0;
		}
		if (e.getSource() == cameraInfo) {
			printCameraStats();
		}
	}

	
	private boolean getNextStepData() throws IOException {
		if (statsRead != null) {
			String line = "";
			while (line != null && line.length() == 0) {
				line = statsRead.readLine();
			}
			if (line != null) {
				// parse line to get new camera heading matrix
				for (int i = 0; i < 3; ++i) {
//					System.out.println("> " + line);
					StringTokenizer tokens = new StringTokenizer(line, " ,");
					int j = -1;
					while (j < 3 && tokens.hasMoreTokens()) {
						camHeading.setElement(i, ++j, Double.parseDouble(tokens.nextToken()));
					}
					line = statsRead.readLine();
				}
//				System.out.println(camHeading + "\n");
//				System.out.println("> " + line);

				// parse line to get new camera postion vector
				StringTokenizer tokens = new StringTokenizer(line, " ,()");
				int j = -1;
				double position[] = new double[3];
				while (j < 3 && tokens.hasMoreTokens()) {
					position[++j] = Double.parseDouble(tokens.nextToken());
				}
				System.out.println(new Vector3d(position));
				camPosition.set(new Vector3d(position));
				
				statsRead.readLine();
			} else
				return false;
		}
		return true;
	}
	

	public void mouseDragged(MouseEvent m) {
		updateCameraTransformation(m.getPoint());
	}

	public void mouseMoved(MouseEvent m) {
		updateCameraTransformation(m.getPoint());
	}

	private void updateCameraTransformation(Point mousePoint){
		Vector2d aim = new Vector2d(mousePoint.x + canvasTopLeft.getX() - refX, mousePoint.y + canvasTopLeft.getY() - refY);
		if (aim.length() < aimTolerance)
			aim.set(0.0, 0.0);
		else
			aim.scale(1 - (aimTolerance / aim.length()));

		// mX = (double) m.getY() - refY;
		mX = aim.y;
		mY = aim.x;

		mX *= Ysign; // Flugsimulator invertMouse
		cosmX = Math.cos(lookSensitivity * mX);
		sinmX = Math.sin(lookSensitivity * mX);
		rotMatX.setColumn(1, 0.0, cosmX, -sinmX);
		rotMatX.setColumn(2, 0.0, sinmX, cosmX);

		// mY = (double) m.getX() - refX;
		cosmY = Math.cos(lookSensitivity * mY);
		sinmY = Math.sin(lookSensitivity * mY);
		rotMatY.setColumn(0, cosmY, 0.0, sinmY);
		rotMatY.setColumn(2, -sinmY, 0.0, cosmY);

		if (!steuerKnueppelMode) { // currently always false
			refY = (double) mousePoint.y; // alt = neu beim Mausdurchdiegegendtragen
			refX = (double) mousePoint.x;
//			try {
//				updateCamera();
//			} catch (IOException ioe) {
//			}
		}

	}
	
	private String printVector(Vector3d vec, int decDig, int fracDig) {
		StringBuffer s = new StringBuffer(15);
		NumberFormat nf = new DecimalFormat();
		nf.setMaximumFractionDigits(fracDig);
		nf.setMinimumFractionDigits(fracDig);
		nf.setMaximumIntegerDigits(decDig);
		nf.setMinimumIntegerDigits(decDig);
		s.append(nf.format(vec.x)).append(' ');
		s.append(nf.format(vec.y)).append(' ');
		s.append(nf.format(vec.z));
		return s.toString();
	}

	public void mouseClicked(MouseEvent m) {
	}

	public void printCameraStats() {
		System.out.println("----------------------------------------------------------------------------");
		System.out.print("Camera heads: \n" + camHeading);
		System.out.println("Camera sits at: " + camPosition);
		System.out.println();
	}

	public void mousePressed(MouseEvent m) { // currently broken
		pickCanvas.setShapeLocation(m);
		PickResult result = pickCanvas.pickClosest();
		if (result == null) {
			System.out.println("Nothing picked");
		} else {
		}

	}

	public void mouseReleased(MouseEvent m) {
	}

	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
	}

	public void mouseWheelMoved(MouseWheelEvent e) {
		mZ = (double) e.getWheelRotation();
		cosmZ = Math.cos(lookSensitivity * mZ * mWheelRollFactor);
		sinmZ = Math.sin(lookSensitivity * mZ * mWheelRollFactor);
		rotMatZ.setColumn(0, cosmZ, sinmZ, 0.0);
		rotMatZ.setColumn(1, -sinmZ, cosmZ, 0.0);

		camHeading.mul(rotMatZ);
	}

	public void setStartPosition(Vector3d startPos) {
		startPosition = new Vector3d(startPos);
	}

	public void start() {
		timer.start();
		resetPosAndHeading();
		setVisible(true);
	}

}